#'
#' Connect and login to an existing Enterprise Server server.
#'
#' @param url Object of class \code{character} representing the Full URL (including schema and port) of the Enterprise Steam server.
#' @param username Object of class \code{character} representing the username.
#' @param password Object of class \code{character} representing user's password or access token.
#' @param verify_ssl (Optional) A \code{logical} value indicating whether to verify SSL when connecting to the server.
#' @param cacert (Optional) Path to a CA bundle file with root and intermediate certificates of trusted CAs.
#' @return This method will connect to the Enterprise Steam server and return a \code{SteamConnection} object that must be passed to subsequent methods.
#' 
#' @examples
#' \dontrun{
#' conn <- h2osteam.login(url = "https://steam.example.com:9555",
#'                        username = "AzureDiamond", password = "hunter2")
#' }
#' 
#' @importFrom RCurl getCurlHandle curlOptions basicHeaderGatherer basicTextGatherer curlPerform
#' @importFrom utils packageVersion
#' @importFrom urltools url_parse url_compose
#' @export
#'
h2osteam.login <- function(url = NA_character_, username = NA_character_, password = NA_character_, verify_ssl = TRUE, cacert = "") {
  # Validate parameter
  if (is.na(url)) {
    stop("Parameter 'url' must be set")
  }
  
  if (is.na(username)) {
    stop("Parameter 'username' must be set")
  }
  
  if (is.na(password)) {
    stop("Parameter 'password' must be set")
  }
  
  # Parse URL
  url <- url_parse(url)
  
  # Clean URL
  url$scheme = "https"
  url$path = NA_character_
  url$parameter = NA_character_
  url$fragment = NA_character_
  if (is.na(url$port)) {
    url$port <- 443
  }
  
  # Save host and port before rebuilding URL
  host <- url$domain
  port <- url$port
  
  # Rebuild URL
  url <- url_compose(url)
  
  # Convert verify_ssl to CURL option
  verifyhost <- 2L
  if (verify_ssl == FALSE) {
    verifyhost <- 0L
  }
  
  # Build CURL options
  curl <- getCurlHandle(cookiefile = "")
  opts <- curlOptions(userpwd = paste(username, password, sep=":"), httpauth = 1L, ssl.verifypeer = verify_ssl, ssl.verifyhost = verifyhost)
  if (cacert != "") {
    opts <- curlOptions(cainfo = cacert, .opts = opts)
  }
  
  # Execute login
  h <- basicHeaderGatherer()
  g <- basicTextGatherer(.mapUnicode = FALSE)
  
  .__curlErr <- FALSE
  .__curlErrMessage <- ""
  tmp <- tryCatch(
    curlPerform(
      url = paste0(url, "auth"),
      writefunction = g$update,
      headerfunction = h$update,
      httpheader = c(
        'User-Agent' = 'Enterprise Steam R Client',
        'Content-Type' = 'application/json'
      ),
      curl = curl,
      .opts = opts
    ),
    error = function(x) { .__curlErr <<- TRUE; .__curlErrMessage <<- x$message }
  )
  
  # Check for errors
  if (.__curlErr) {
    stop(sprintf("%s: %s", .__curlErrMessage, g$value()[1]))
  } else if (h$value()["status"] != 200 && h$value()["status"] != 307){
    stop(g$value()[1])
  }
  
  # Extract cookie
  setCookie <- h$value()["set-cookie"]
  cookie <- strsplit(setCookie, ";")$`set-cookie`[1]
  
  if (is.null(cookie)) {
    setCookie <- h$value()["Set-Cookie"]
    cookie <- strsplit(setCookie, ";")$`Set-Cookie`[1]
  }
  
  if (is.null(cookie)) {
    setCookie <- h$value()["SET-COOKIE"]
    cookie <- strsplit(setCookie, ";")$`SET-COOKIE`[1]
  }

  # Create conn object
  conn <- structure(
    list(url=url, host=host, port=port, verify_ssl=verify_ssl, cacert=cacert, username=username, password=password, cookie=cookie, uid=0, curl=curl),
    class="SteamConnection"
  )

  # Test connection
  client_api_version <- packageVersion("h2osteam")
  server_api_version <- .pingServer(conn, "R connect")
  if (server_api_version != client_api_version) {
      stop(sprintf("Client API version '%s' does not match server API version '%s'", client_api_version, server_api_version))
  }

  conn
}